# Copyright (c) 2018 James Patrick Dill
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import asyncio
from functools import wraps


def run_every(s, cancel_old_task=False):
    """
    This decorator will run a function every `s` seconds after the on_ready event.

    :param int s: seconds to wait
    :param bool cancel_old_task: (keyword) If True, the previous task will be canceled if still running.
    """

    def real_decorator(func):

        @wraps(func)
        async def new_func(self, loop):
            task = loop.create_task(func(self))  # create task

            while True:
                await asyncio.sleep(s)

                if cancel_old_task and not task.done():
                    task.cancel()  # if necessary, cancel still running task

                task = loop.create_task(func(self))  # create task again

        new_func.is_bg_task = True

        return new_func

    return real_decorator



# Copyright (c) 2018 James Patrick Dill
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import re

from outlet import errors

mention = re.compile(r"<@!?([0-9]+)>")
username = re.compile("(.+#[0-9]{4})")


class Converter(object):
    """
    A converter converts a raw argument to a command to a usable value.

    Custom converters should inherit outlet.BaseConverter, and must have a convert(value, ctx) classmethod.
    If the value can't be converted, it should raise outlet.errors.WrongType
    """

    @classmethod
    def convert(cls, value, ctx):
        return


# basic converters


class String(Converter):
    """String converter. Just returns the raw value."""

    @classmethod
    def convert(cls, value, ctx):
        return value


class Number(Converter):
    """Number converter. Will return int or float"""

    @classmethod
    def convert(cls, value, ctx):
        try:
            return float(value) if "." in value else int(value)
        except ValueError:
            raise errors.WrongType("`{}` isn't a number.".format(value))


class Member(Converter):
    """
    Member converter. Converts a mention to :class:`discord.Member` . The mention must be for a member of the guild.

    This converter works with @mentions or username#1234
    """

    @classmethod
    def convert(cls, value, ctx):
        user_id = mention.match(value)
        user_tag = username.match(value)

        if not (user_tag or user_id):
            raise errors.WrongType("`{}` isn't a username or mention.".format(value))

        if user_id:
            user_id = int(user_id.group(1))
            member = ctx.guild.get_member(user_id)

            if member is None:
                raise errors.WrongType("{} isn't a member of this server".format(value))

            return member
        else:
            member = ctx.guild.get_member_named(value)
            if member is None:
                raise errors.WrongType("{} isn't a member of this server".format(value))

            return member


class RelativeTime(Converter):
    """
    Converts a relative time to its number of seconds (e.g., 10:45 = 10*60 + 45 = 645)
    """

    @classmethod
    def convert(cls, value, ctx):
        try:
            times = [int(n) for n in value.split(":")]
        except ValueError:
            raise errors.WrongType("Times should be in the format `hh:mm:ss`")
        else:
            times = [0] * (3 - len(times)) + times

            return times[0] * 60 * 60 + times[1] * 60 + times[2]



# Copyright (c) 2018 James Patrick Dill
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from discord import TextChannel
from outlet import errors


class Plugin(object):
    """
    Plugin class. Create your own plugins by inheriting this class.

    Events
    ======

    To use the raw events, you can overwrite them like a regular discord client. For example ::

        class Plugin(outlet.Plugin):
            async def on_ready(self):
                print("I'm ready!")



    The Plugin class *must* be named Plugin for it to be used.

    Commands
    ========

    You can also create commands with the command decorator.
    Commands are passed a :class:`Context`, which has information about the message and helper functions.
    For example ::
        @outlet.command("ping")
        async def ping_pong(ctx):
            ctx.send("pong")


    """

    #: Name of plugin.
    __plugin__ = "default"

    def __init__(self, bot):
        #: Bot the plugin belongs to
        self.bot = bot

        # clone bot
        self.get_resource = self.bot.get_resource

        self.http = self.bot.http_session

        self.log = self.bot.log

        self.commands = {}
        self.event_listeners = {
            "on_message": []
        }
        self.bg_tasks = []

        self.running_bg_tasks = []

        self.collect_commands()
        self.collect_event_listeners()
        self.collect_bg_tasks()

    def collect_commands(self):
        # collects commands

        self.log.debug("getting commands")

        for name in dir(self):  # get commands
            o = getattr(self, name)
            if getattr(o, "is_command", False):
                for cmd in o.command:
                    self.commands[cmd] = o

        self.log.info("loaded {} commands".format(len(self.commands)))

    def collect_event_listeners(self):
        # collects event listeners

        self.log.debug("getting event listeners")

        for name in dir(self):
            o = getattr(self, name)
            if getattr(o, "is_event_listener", False):
                self.event_listeners[o.event].append(o)

        listeners = sum([len(e) for e in self.event_listeners.values()])

        self.log.info("loaded {} event listeners".format(listeners))

    def collect_bg_tasks(self):
        # collects background tasks

        self.log.debug("getting event listeners")

        for name in dir(self):
            o = getattr(self, name)
            if getattr(o, "is_bg_task", False):
                self.bg_tasks.append(o)

        self.log.info("loaded {} background tasks".format(len(self.bg_tasks)))

    def create_task(self, *args, **kwargs):
        """
        Shortcut to :meth:`Plugin.bot.loop.create_task`

        Call this on a coroutine to run it without blocking.

        :returns: asyncio.Task
        """

        return self.bot.loop.create_task(*args, **kwargs)

    async def start_bg_tasks(self):
        self.log.info("cancelling running background tasks")
        for bg_task in self.running_bg_tasks:  # first cancel the running tasks
            if not bg_task.done():
                bg_task.cancel()

        self.log.info("starting background tasks")
        for bg_task in self.bg_tasks:
            self.create_task(bg_task(self.bot.loop))  # create the task, use the bot's event loop

            self.log.debug("started {}".format(bg_task.__name__))

    # bot level event handling

    async def __on_message__(self, message):
        self.create_task(self.on_message(message))

        if isinstance(message.channel, TextChannel):
            for event_listener in self.event_listeners["on_message"]:
                if event_listener.channel is not None and message.channel.name != event_listener.channel:
                    continue

                self.create_task(event_listener(message))

        # command handling
        if message.content.startswith(self.bot.prefix) and message.content != self.bot.prefix:
            self.log.info("command received")

            if message.author == self.bot.user:  # ignore myself!
                return

            no_prefix = message.content[len(self.bot.prefix):]
            cmd = no_prefix.split(" ")[0]

            if cmd in self.commands:
                try:
                    await self.commands[cmd](message, no_prefix)
                except errors.CommandError as e:
                    self.log.info("command raised {}: {!s}".format(e.__class__.__name__, e))

                    await message.channel.send(e)

    async def __on_ready__(self):
        self.create_task(self.start_bg_tasks())

        self.create_task(self.on_ready())

    # plugin level event handling

    async def on_ready(self):
        pass

    async def on_shard_ready(self, shard_id):
        pass

    async def on_typing(self, channel, user, when):
        pass

    async def on_message(self, message):
        pass

    async def on_message_delete(self, message):
        pass

    async def on_message_edit(self, before, after):
        pass

    async def on_reaction_add(self, reaction, user):
        pass

    async def on_reaction_remove(self, reaction, user):
        pass

    async def on_reaction_clear(self, message, reactions):
        pass

    async def on_private_channel_delete(self, channel):
        pass

    async def on_private_channel_create(self, channel):
        pass

    async def on_private_channel_update(self, before, after):
        pass

    async def on_guild_channel_create(self, channel):
        pass

    async def on_guild_channel_delete(self, channel):
        pass

    async def on_guild_channel_update(self, before, after):
        pass

    async def on_member_join(self, member):
        pass

    async def on_member_remove(self, member):
        pass

    async def on_member_update(self, before, after):
        pass

    async def on_guild_join(self, guild):
        pass

    async def on_guild_remove(self, guild):
        pass

    async def on_guild_update(self, before, after):
        pass

    async def on_guild_role_create(self, role):
        pass

    async def on_guild_role_delete(self, role):
        pass

    async def on_guild_role_update(self, before, after):
        pass

    async def on_guild_emojis_update(self, guild, before, after):
        pass

    async def on_guild_available(self, guild):
        pass

    async def on_guild_unavailable(self, guild):
        pass

    async def on_member_ban(self, guild, user):
        pass

    async def on_member_unban(self, guild, user):
        pass



# Copyright (c) 2018 James Patrick Dill
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import time
from functools import wraps
from inspect import Signature, Parameter

from outlet import errors, converters


class Context(object):
    """
    Command context, passed to command functions for easier handling

    :attr discord.Message message: Message
    :attr discord.Guild guild: Guild the message was sent in
    :attr discord.Channel channel: Channel the message was sent in
    :attr discord.Member: author: Author of the message
    """

    def __init__(self, message):
        self.message = message

        self.guild = message.guild
        self.channel = message.channel
        self.author = message.author

    async def send(self, *args, **kwargs):
        """
        Coroutine

        Sends a message in the context's channel. Pass the same arguments or keywords that you would to
        :meth:`discord.Channel.send`
        """

        await self.channel.send(*args, **kwargs)


def convert_arguments(args, signature, ctx):
    """
    This is used internally by the bot.

    Converts a list of arguments for a command against the command's signature using converters.

    :param list args: Arguments
    :param Signature signature:  Function signature
    :param Context ctx: Command context
    :return: list
    """

    if len(signature.parameters) == 2:  # command doesn't take any arguments
        return []

    params = list(signature.parameters.values())[2:]  # convert parameters to list, ignore self and ctx

    has_var_positional = params[-1].kind == Parameter.VAR_POSITIONAL  # has *arg

    min_args = len(params) - 1 if has_var_positional else len(params)  # *arg not required
    max_args = 999 if has_var_positional else len(params)

    if len(args) < min_args:
        raise errors.MissingArguments("{} argument(s) are required for this command".format(min_args))
    elif len(args) > max_args:
        raise errors.TooManyArguments("Only {} argument(s) are allowed for this command".format(max_args))

    converted_args = []

    for i in range(len(params)):
        if len(args) == 0:
            break

        while args[0] == "":  # ignore empty arguments caused by double spaces
            del args[0]

        converter = params[i].annotation

        if converter == Parameter.empty:  # use String by default
            converter = converters.String

        # if not issubclass(type(converter), converters.Converter):  # check that it's actually a converter
        #     raise TypeError("all converters should be a subclass of outlet.converters.Converter")

        new_value = converter.convert(args[0], ctx)
        converted_args.append(new_value)

        del args[0]

        if params[i].kind == Parameter.VAR_POSITIONAL:  # if *arg, convert the rest
            while len(args) > 0:
                new_value = converter.convert(args[0], ctx)
                converted_args.append(new_value)

                del args[0]

    return converted_args


def help_message(sig, doc):
    msg = ""

    params = list(sig.parameters.values())
    params = params[2:] if len(params) > 2 else []

    for arg in params:
        type_ = "String" if arg.annotation == Parameter.empty else arg.annotation.__name__

        msg += "{}<{}: {}> ".format("*" if arg.kind == Parameter.VAR_POSITIONAL else "", arg.name, type_)

    return "{}\n\n{}".format(msg, doc or "")


def command(*cmd):
    """
    Decorator used to create commands.

    :param *cmd: The name(s) of the command, used to call it. Don't include the bot prefix.
    """

    def real_decorator(func):
        signature = Signature.from_callable(func)  # function signature is used to convert arguments

        @wraps(func)
        async def real_command(self_, message, content):
            content = content.split(" ")  # the content parameter doesn't include the command prefix

            command_ = content[0]  # actual named command
            args = content[1:] if len(content) > 1 else []  # list of arguments passed

            if command_ not in cmd:  # doesn't match this command
                return

            ctx = Context(message)  # message context is passed to function

            args = convert_arguments(args, signature, ctx)  # convert the arguments according to function signature
            # this will also checked for incorrect number of arguments

            returned = await func(self_, ctx, *args)

            # automatically send return value to channel if its a string
            if isinstance(returned, str):
                await ctx.send(returned)

        real_command.is_command = True
        real_command.command = cmd
        real_command.help_message = help_message(signature, func.__doc__)

        return real_command

    return real_decorator


# helpers

def require_permissions(*permission):
    """
    Decorator that makes command require named permission(s)

    :param str permission: permission(s) to require
    """

    def real_decorator(func):
        if getattr(func, "is_command", False):
            # @outlet.command("h")
            # @require_permissions("h")
            # async def h_command(self, ctx):

            raise SyntaxError("@require_permissions() decorator should be placed under the @command decorator")

        @wraps(func)
        async def new_func(self_, ctx, *args):
            author_permissions = ctx.author.permissions_in(ctx.channel)  # use per channel permissions

            for perm in permission:  # check all permissions
                if not getattr(author_permissions, perm, False):
                    raise errors.MissingPermission("This command requires the `{}` permission".format(perm))

            return await func(self_, ctx, *args)

        return new_func

    return real_decorator


def owner_only(func):
    """
    Decorator that only allows the owner of the guild to use command.
    """

    if getattr(func, "is_command", False):
        raise SyntaxError("@owner_only decorator should be placed under the @command decorator")

    @wraps(func)
    async def new_func(self_, ctx, *args):
        if ctx.author != ctx.guild.owner:
            raise errors.MissingPermission("Only the owner of the guild can use this command")

        return await func(self_, ctx, *args)

    return new_func


def cooldown(seconds):
    """
    Adds a per-server cooldown to a command.

    :param int seconds: Length of cooldown in seconds.
    :return:
    """

    def real_decorator(func):
        if getattr(func, "is_command", False):
            # @outlet.command("h")
            # @require_permissions("h")
            # async def h_command(self, ctx):

            raise SyntaxError("@require_permissions() decorator should be placed under the @command decorator")

        guild_calls = {}

        @wraps(func)
        async def new_func(self_, ctx, *args):
            if time.time() - guild_calls.get(ctx.guild.id, 0) < seconds:
                raise errors.CommandError("This command has a {} second cooldown.".format(seconds))

            guild_calls[ctx.guild.id] = time.time()

            return await func(self_, ctx, *args)

        return new_func

    return real_decorator



import outlet
from outlet import errors

import discord

from inspect import signature, Parameter


class Plugin(outlet.Plugin):
    """
    Builtin help plugin, implementing $help command
    """

    __plugin__ = "Help"

    async def help_embed(self, guild):
        new_embed = discord.Embed(name="Commands",
                                  description="Do **$help <command>** for more information about a command.",
                                  color=await self.bot.my_color(guild))

        new_embed.set_author(name="Commands")

        for plugin in self.bot.plugins:
            commands = []
            for command in plugin.commands.values():
                for cmd_name in command.command:
                    commands.append(self.bot.prefix + cmd_name)

            command_list = "\n".join(commands)

            if commands:
                new_embed.add_field(name=plugin.__plugin__, value=command_list)

        return new_embed

    @outlet.command("help")
    async def help(self, ctx, *command):
        """Shows a list of commands, or help for a specific command."""

        if command:
            command = command[0]  # ignore any lingering stuff

            if command.startswith(self.bot.prefix) and command != self.bot.prefix:
                command = command[len(self.bot.prefix):]

            cmd = self.bot.all_commands.get(command)

            if cmd is None:
                raise errors.ArgumentError("`{}` isn't a command".format(command))

            await ctx.send("{}**{}** {}".format(self.bot.prefix, command, cmd.help_message))


        else:
            embed = await self.help_embed(ctx.guild)

            await ctx.send(embed=embed)



# Copyright (c) 2018 James Patrick Dill
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import glob
import logging
import os

import discord

import aiohttp

from outlet import builtin_plugins, util

log = logging.getLogger("outlet")


class DiscordBot(discord.Client):
    """
    Bot class. All of the code for your bot is written in plugins, this class simply collects all the plugins and
    funnels the events into them.

    :param str token: Bot token used to connect to Discord
    :keyword str plugin_dir: Directory where plugin files are stored.
    :keyword str prefix: Bot prefix. Defaults to "!"
    :keyword logger: Logger to be used by bot. If none is given, Outlet's logger is used.
    """

    def __init__(self, token, *, plugin_dir, resource_dir, prefix="!", logger=log):
        super().__init__()

        self.token = token

        self.plugin_dir = plugin_dir
        self.resource_dir = resource_dir

        self.http_session = aiohttp.ClientSession()

        self.prefix = prefix

        self.log = logger

        self.all_commands = {}

        self.plugins = self.get_plugins(plugin_dir)
        self.get_all_commands()

    def get_resource(self, file):
        return util.import_file(os.path.join(self.resource_dir, file))

    def run(self):
        self.log.info("starting bot")

        super().run(self.token)

    def get_all_commands(self):
        # get dict with all commands for help plugin

        for plugin in self.plugins:
            self.all_commands.update(plugin.commands)

    def get_plugins(self, plugin_dir):
        # import all plugins from directory

        plugins = [plugin(self) for plugin in builtin_plugins.plugins]

        files = glob.glob(os.path.join(plugin_dir, "*.py"))
        for file in files:
            try:
                self.log.info("importing {}".format(file))

                plugins.append(
                    util.import_file(file).Plugin(self)  # import file and get plugin
                )

            except ImportError as e:
                self.log.error("failed to load {}\n\n{}".format(file, e))

        return plugins

    # # log errors
    #
    # async def on_error(self, event_method, *args, **kwargs):
    #     exc = sys.exc_info()[1]
    #
    #     self.log.error("exception while handling {} event: {!s}".format(event_method, exc))

    # util functions

    async def my_color(self, guild):
        """
        Returns the color of the bot's role in a specific guild.

        :param guild: Guild to get color from.
        """

        me = guild.get_member(self.user.id)
        if me is None:
            raise ValueError("bot must be in guild to get its color")

        return me.color

    # event funneling

    async def on_ready(self):
        self.log.debug("bot ready")

        for plugin in self.plugins:
            await plugin.__on_ready__()

    async def on_shard_ready(self, shard_id):
        self.log.debug("shard {} ready".format(shard_id))

        for plugin in self.plugins:
            await plugin.on_shard_ready()

    async def on_typing(self, channel, user, when):
        for plugin in self.plugins:
            await plugin.on_typing(channel, user, when)

    async def on_message(self, message):
        for plugin in self.plugins:
            await plugin.__on_message__(message)

    async def on_message_delete(self, message):
        for plugin in self.plugins:
            await plugin.on_message_delete(message)

    async def on_message_edit(self, before, after):
        for plugin in self.plugins:
            await plugin.on_message_edit(before, after)

    async def on_reaction_add(self, reaction, user):
        for plugin in self.plugins:
            await plugin.on_reaction_add(reaction, user)

    async def on_reaction_remove(self, reaction, user):
        for plugin in self.plugins:
            await plugin.on_reaction_remove(reaction, user)

    async def on_reaction_clear(self, message, reactions):
        for plugin in self.plugins:
            await plugin.on_reaction_clear(message, reactions)

    async def on_private_channel_delete(self, channel):
        for plugin in self.plugins:
            await plugin.on_private_channel_delete(channel)

    async def on_private_channel_create(self, channel):
        for plugin in self.plugins:
            await plugin.on_private_channel_create(channel)

    async def on_private_channel_update(self, before, after):
        for plugin in self.plugins:
            await plugin.on_private_channel_update(before, after)

    async def on_guild_channel_create(self, channel):
        for plugin in self.plugins:
            await plugin.on_guild_channel_create(channel)

    async def on_guild_channel_delete(self, channel):
        for plugin in self.plugins:
            await plugin.on_guild_channel_delete(channel)

    async def on_guild_channel_update(self, before, after):
        for plugin in self.plugins:
            await plugin.on_guild_channel_update(before, after)

    async def on_member_join(self, member):
        for plugin in self.plugins:
            await plugin.on_member_join(member)

    async def on_member_remove(self, member):
        for plugin in self.plugins:
            await plugin.on_member_remove(member)

    async def on_member_update(self, before, after):
        for plugin in self.plugins:
            await plugin.on_member_update(before, after)

    async def on_guild_join(self, guild):
        for plugin in self.plugins:
            await plugin.on_guild_join(guild)

    async def on_guild_remove(self, guild):
        for plugin in self.plugins:
            await plugin.on_guild_remove(guild)

    async def on_guild_update(self, before, after):
        for plugin in self.plugins:
            await plugin.on_guild_update(before, after)

    async def on_guild_role_create(self, role):
        for plugin in self.plugins:
            await plugin.on_guild_role_create(role)

    async def on_guild_role_delete(self, role):
        for plugin in self.plugins:
            await plugin.on_guild_role_delete(role)

    async def on_guild_role_update(self, before, after):
        for plugin in self.plugins:
            await plugin.on_guild_role_update(before, after)

    async def on_guild_emojis_update(self, guild, before, after):
        for plugin in self.plugins:
            await plugin.on_guild_emojis_update(guild, before, after)

    async def on_guild_available(self, guild):
        for plugin in self.plugins:
            await plugin.on_guild_available(guild)

    async def on_guild_unavailable(self, guild):
        for plugin in self.plugins:
            await plugin.on_guild_unavailable(guild)

    async def on_member_ban(self, guild, user):
        for plugin in self.plugins:
            await plugin.on_member_ban(guild, user)

    async def on_member_unban(self, guild, user):
        for plugin in self.plugins:
            await plugin.on_member_unban(guild, user)



class CommandError(Exception):
    pass


class MissingPermission(CommandError):
    pass


class ArgumentError(CommandError):
    pass


class MissingArguments(ArgumentError):
    pass


class TooManyArguments(ArgumentError):
    pass


class WrongType(ArgumentError):
    pass

